@@ -28,7 +28,7 @@
 #include "lifecycle_msgs/msg/state.hpp"
 #include "rclcpp/rclcpp.hpp"
 
-using CallbackReturn = rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn;
+using CallbackReturn = controller_interface::CallbackReturn;
 using hardware_interface::HW_IF_POSITION;
 using hardware_interface::HW_IF_VELOCITY;
 using hardware_interface::LoanedCommandInterface;
@@ -48,12 +48,12 @@ public:
   }
 
   /**
-  * @brief wait_for_twist block until a new twist is received.
-  * Requires that the executor is not spinned elsewhere between the
-  *  message publication and the call to this function
-  *
-  * @return true if new twist msg was received, false if timeout
-  */
+   * @brief wait_for_twist block until a new twist is received.
+   * Requires that the executor is not spinned elsewhere between the
+   *  message publication and the call to this function
+   *
+   * @return true if new twist msg was received, false if timeout
+   */
   bool wait_for_twist(
     rclcpp::Executor & executor,
     const std::chrono::milliseconds & timeout = std::chrono::milliseconds(500))
@@ -68,6 +68,17 @@ public:
     }
     return false;
   }
+
+  /**
+   * @brief Used to get the real_time_odometry_publisher to verify its contents
+   *
+   * @return Copy of realtime_odometry_publisher_ object
+   */
+  std::shared_ptr<realtime_tools::RealtimePublisher<nav_msgs::msg::Odometry>>
+  get_rt_odom_publisher()
+  {
+    return realtime_odometry_publisher_;
+  }
 };
 
 class TestDiffDriveController : public ::testing::Test
@@ -244,6 +255,216 @@ TEST_F(TestDiffDriveController, configure_succeeds_when_wheels_are_specified)
     SizeIs(left_wheel_names.size() + right_wheel_names.size()));
 }
 
+TEST_F(TestDiffDriveController, configure_succeeds_tf_test_prefix_false_no_namespace)
+{
+  const auto ret = controller_->init(controller_name);
+  ASSERT_EQ(ret, controller_interface::return_type::OK);
+
+  std::string odom_id = "odom";
+  std::string base_link_id = "base_link";
+  std::string frame_prefix = "test_prefix";
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("left_wheel_names", rclcpp::ParameterValue(left_wheel_names)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("right_wheel_names", rclcpp::ParameterValue(right_wheel_names)));
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix_enable", rclcpp::ParameterValue(false)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix", rclcpp::ParameterValue(frame_prefix)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("odom_frame_id", rclcpp::ParameterValue(odom_id)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("base_frame_id", rclcpp::ParameterValue(base_link_id)));
+
+  ASSERT_EQ(controller_->on_configure(rclcpp_lifecycle::State()), CallbackReturn::SUCCESS);
+
+  auto odometry_message = controller_->get_rt_odom_publisher()->msg_;
+  std::string test_odom_frame_id = odometry_message.header.frame_id;
+  std::string test_base_frame_id = odometry_message.child_frame_id;
+  /* tf_frame_prefix_enable is false so no modifications to the frame id's */
+  ASSERT_EQ(test_odom_frame_id, odom_id);
+  ASSERT_EQ(test_base_frame_id, base_link_id);
+}
+
+TEST_F(TestDiffDriveController, configure_succeeds_tf_test_prefix_true_no_namespace)
+{
+  const auto ret = controller_->init(controller_name);
+  ASSERT_EQ(ret, controller_interface::return_type::OK);
+
+  std::string odom_id = "odom";
+  std::string base_link_id = "base_link";
+  std::string frame_prefix = "test_prefix";
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("left_wheel_names", rclcpp::ParameterValue(left_wheel_names)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("right_wheel_names", rclcpp::ParameterValue(right_wheel_names)));
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix_enable", rclcpp::ParameterValue(true)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix", rclcpp::ParameterValue(frame_prefix)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("odom_frame_id", rclcpp::ParameterValue(odom_id)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("base_frame_id", rclcpp::ParameterValue(base_link_id)));
+
+  ASSERT_EQ(controller_->on_configure(rclcpp_lifecycle::State()), CallbackReturn::SUCCESS);
+
+  auto odometry_message = controller_->get_rt_odom_publisher()->msg_;
+  std::string test_odom_frame_id = odometry_message.header.frame_id;
+  std::string test_base_frame_id = odometry_message.child_frame_id;
+
+  /* tf_frame_prefix_enable is true and frame_prefix is not blank so should be appended to the frame
+   * id's */
+  ASSERT_EQ(test_odom_frame_id, frame_prefix + "/" + odom_id);
+  ASSERT_EQ(test_base_frame_id, frame_prefix + "/" + base_link_id);
+}
+
+TEST_F(TestDiffDriveController, configure_succeeds_tf_blank_prefix_true_no_namespace)
+{
+  const auto ret = controller_->init(controller_name);
+  ASSERT_EQ(ret, controller_interface::return_type::OK);
+
+  std::string odom_id = "odom";
+  std::string base_link_id = "base_link";
+  std::string frame_prefix = "";
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("left_wheel_names", rclcpp::ParameterValue(left_wheel_names)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("right_wheel_names", rclcpp::ParameterValue(right_wheel_names)));
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix_enable", rclcpp::ParameterValue(true)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix", rclcpp::ParameterValue(frame_prefix)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("odom_frame_id", rclcpp::ParameterValue(odom_id)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("base_frame_id", rclcpp::ParameterValue(base_link_id)));
+
+  ASSERT_EQ(controller_->on_configure(rclcpp_lifecycle::State()), CallbackReturn::SUCCESS);
+
+  auto odometry_message = controller_->get_rt_odom_publisher()->msg_;
+  std::string test_odom_frame_id = odometry_message.header.frame_id;
+  std::string test_base_frame_id = odometry_message.child_frame_id;
+  /* tf_frame_prefix_enable is true but frame_prefix is blank so should not be appended to the frame
+   * id's */
+  ASSERT_EQ(test_odom_frame_id, odom_id);
+  ASSERT_EQ(test_base_frame_id, base_link_id);
+}
+
+TEST_F(TestDiffDriveController, configure_succeeds_tf_test_prefix_false_set_namespace)
+{
+  std::string test_namespace = "/test_namespace";
+
+  const auto ret = controller_->init(controller_name, test_namespace);
+  ASSERT_EQ(ret, controller_interface::return_type::OK);
+
+  std::string odom_id = "odom";
+  std::string base_link_id = "base_link";
+  std::string frame_prefix = "test_prefix";
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("left_wheel_names", rclcpp::ParameterValue(left_wheel_names)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("right_wheel_names", rclcpp::ParameterValue(right_wheel_names)));
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix_enable", rclcpp::ParameterValue(false)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix", rclcpp::ParameterValue(frame_prefix)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("odom_frame_id", rclcpp::ParameterValue(odom_id)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("base_frame_id", rclcpp::ParameterValue(base_link_id)));
+
+  ASSERT_EQ(controller_->on_configure(rclcpp_lifecycle::State()), CallbackReturn::SUCCESS);
+
+  auto odometry_message = controller_->get_rt_odom_publisher()->msg_;
+  std::string test_odom_frame_id = odometry_message.header.frame_id;
+  std::string test_base_frame_id = odometry_message.child_frame_id;
+  /* tf_frame_prefix_enable is false so no modifications to the frame id's */
+  ASSERT_EQ(test_odom_frame_id, odom_id);
+  ASSERT_EQ(test_base_frame_id, base_link_id);
+}
+
+TEST_F(TestDiffDriveController, configure_succeeds_tf_test_prefix_true_set_namespace)
+{
+  std::string test_namespace = "/test_namespace";
+
+  const auto ret = controller_->init(controller_name, test_namespace);
+  ASSERT_EQ(ret, controller_interface::return_type::OK);
+
+  std::string odom_id = "odom";
+  std::string base_link_id = "base_link";
+  std::string frame_prefix = "test_prefix";
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("left_wheel_names", rclcpp::ParameterValue(left_wheel_names)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("right_wheel_names", rclcpp::ParameterValue(right_wheel_names)));
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix_enable", rclcpp::ParameterValue(true)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix", rclcpp::ParameterValue(frame_prefix)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("odom_frame_id", rclcpp::ParameterValue(odom_id)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("base_frame_id", rclcpp::ParameterValue(base_link_id)));
+
+  ASSERT_EQ(controller_->on_configure(rclcpp_lifecycle::State()), CallbackReturn::SUCCESS);
+
+  auto odometry_message = controller_->get_rt_odom_publisher()->msg_;
+  std::string test_odom_frame_id = odometry_message.header.frame_id;
+  std::string test_base_frame_id = odometry_message.child_frame_id;
+
+  /* tf_frame_prefix_enable is true and frame_prefix is not blank so should be appended to the frame
+   * id's instead of the namespace*/
+  ASSERT_EQ(test_odom_frame_id, frame_prefix + "/" + odom_id);
+  ASSERT_EQ(test_base_frame_id, frame_prefix + "/" + base_link_id);
+}
+
+TEST_F(TestDiffDriveController, configure_succeeds_tf_blank_prefix_true_set_namespace)
+{
+  std::string test_namespace = "/test_namespace";
+
+  const auto ret = controller_->init(controller_name, test_namespace);
+  ASSERT_EQ(ret, controller_interface::return_type::OK);
+
+  std::string odom_id = "odom";
+  std::string base_link_id = "base_link";
+  std::string frame_prefix = "";
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("left_wheel_names", rclcpp::ParameterValue(left_wheel_names)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("right_wheel_names", rclcpp::ParameterValue(right_wheel_names)));
+
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix_enable", rclcpp::ParameterValue(true)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("tf_frame_prefix", rclcpp::ParameterValue(frame_prefix)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("odom_frame_id", rclcpp::ParameterValue(odom_id)));
+  controller_->get_node()->set_parameter(
+    rclcpp::Parameter("base_frame_id", rclcpp::ParameterValue(base_link_id)));
+
+  ASSERT_EQ(controller_->on_configure(rclcpp_lifecycle::State()), CallbackReturn::SUCCESS);
+
+  auto odometry_message = controller_->get_rt_odom_publisher()->msg_;
+  std::string test_odom_frame_id = odometry_message.header.frame_id;
+  std::string test_base_frame_id = odometry_message.child_frame_id;
+  /* tf_frame_prefix_enable is true but frame_prefix is blank so namespace should be appended to the
+   * frame id's */
+  ASSERT_EQ(test_odom_frame_id, test_namespace + "/" + odom_id);
+  ASSERT_EQ(test_base_frame_id, test_namespace + "/" + base_link_id);
+}
+
 TEST_F(TestDiffDriveController, activate_fails_without_resources_assigned)
 {
   const auto ret = controller_->init(controller_name);
@@ -339,11 +560,11 @@ TEST_F(TestDiffDriveController, cleanup)
 
   rclcpp::executors::SingleThreadedExecutor executor;
   executor.add_node(controller_->get_node()->get_node_base_interface());
-  auto state = controller_->configure();
+  auto state = controller_->get_node()->configure();
   ASSERT_EQ(State::PRIMARY_STATE_INACTIVE, state.id());
   assignResourcesPosFeedback();
 
-  state = controller_->activate();
+  state = controller_->get_node()->activate();
   ASSERT_EQ(State::PRIMARY_STATE_ACTIVE, state.id());
 
   waitForSetup();
@@ -358,13 +579,13 @@ TEST_F(TestDiffDriveController, cleanup)
     controller_->update(rclcpp::Time(0, 0, RCL_ROS_TIME), rclcpp::Duration::from_seconds(0.01)),
     controller_interface::return_type::OK);
 
-  state = controller_->deactivate();
+  state = controller_->get_node()->deactivate();
   ASSERT_EQ(State::PRIMARY_STATE_INACTIVE, state.id());
   ASSERT_EQ(
     controller_->update(rclcpp::Time(0, 0, RCL_ROS_TIME), rclcpp::Duration::from_seconds(0.01)),
     controller_interface::return_type::OK);
 
-  state = controller_->cleanup();
+  state = controller_->get_node()->cleanup();
   ASSERT_EQ(State::PRIMARY_STATE_UNCONFIGURED, state.id());
 
   // should be stopped
@@ -389,14 +610,14 @@ TEST_F(TestDiffDriveController, correct_initialization_using_parameters)
   rclcpp::executors::SingleThreadedExecutor executor;
   executor.add_node(controller_->get_node()->get_node_base_interface());
 
-  auto state = controller_->configure();
+  auto state = controller_->get_node()->configure();
   assignResourcesPosFeedback();
 
   ASSERT_EQ(State::PRIMARY_STATE_INACTIVE, state.id());
   EXPECT_EQ(0.01, left_wheel_vel_cmd_.get_value());
   EXPECT_EQ(0.02, right_wheel_vel_cmd_.get_value());
 
-  state = controller_->activate();
+  state = controller_->get_node()->activate();
   ASSERT_EQ(State::PRIMARY_STATE_ACTIVE, state.id());
 
   // send msg
@@ -415,7 +636,7 @@ TEST_F(TestDiffDriveController, correct_initialization_using_parameters)
   // deactivated
   // wait so controller process the second point when deactivated
   std::this_thread::sleep_for(std::chrono::milliseconds(500));
-  state = controller_->deactivate();
+  state = controller_->get_node()->deactivate();
   ASSERT_EQ(state.id(), State::PRIMARY_STATE_INACTIVE);
   ASSERT_EQ(
     controller_->update(rclcpp::Time(0, 0, RCL_ROS_TIME), rclcpp::Duration::from_seconds(0.01)),
@@ -425,12 +646,12 @@ TEST_F(TestDiffDriveController, correct_initialization_using_parameters)
   EXPECT_EQ(0.0, right_wheel_vel_cmd_.get_value()) << "Wheels are halted on deactivate()";
 
   // cleanup
-  state = controller_->cleanup();
+  state = controller_->get_node()->cleanup();
   ASSERT_EQ(State::PRIMARY_STATE_UNCONFIGURED, state.id());
   EXPECT_EQ(0.0, left_wheel_vel_cmd_.get_value());
   EXPECT_EQ(0.0, right_wheel_vel_cmd_.get_value());
 
-  state = controller_->configure();
+  state = controller_->get_node()->configure();
   ASSERT_EQ(State::PRIMARY_STATE_INACTIVE, state.id());
   executor.cancel();
 }
